# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from . import utils
from ..data import get_origins


def test_invalid_get(client):
    query_str = """
    query getOrigin {
      origin(url: "http://example.com/non-existing") {
        url
      }
    }
    """
    utils.assert_missing_object(client, query_str, "origin")


@pytest.mark.parametrize("origin", get_origins())
def test_get(client, storage, origin):
    query_str = """
    query getOrigin($url: String!) {
      origin(url: $url) {
        url
        id
        visits(first: 10) {
          nodes {
            id
          }
        }
        latestVisit {
          visitId
        }
        snapshots(first: 2) {
          nodes {
            id
          }
        }
      }
    }
    """

    response, _ = utils.get_query_response(client, query_str, url=origin.url)
    data_origin = response["origin"]
    storage_origin = storage.origin_get([origin.url])[0]
    visits_and_statuses = storage.origin_visit_get_with_statuses(origin.url).results
    assert data_origin["url"] == storage_origin.url
    assert data_origin["id"] == storage_origin.id.hex()
    assert len(data_origin["visits"]["nodes"]) == len(visits_and_statuses)
    assert data_origin["latestVisit"]["visitId"] == visits_and_statuses[-1].visit.visit
    snapshots = storage.origin_snapshot_get_all(origin.url)
    assert len(data_origin["snapshots"]["nodes"]) == len(snapshots)


@pytest.mark.parametrize("origin", get_origins())
@pytest.mark.parametrize("sort", ["ASC", "DESC"])
def test_visits_sort_order(client, storage, origin, sort):
    query_str = """
    query getOrigin($url: String!, $sort: ListOrder) {
      origin(url: $url) {
        url
        id
        visits(first: 10, sort: $sort) {
          nodes {
            visitId
          }
        }
      }
    }
    """
    response, _ = utils.get_query_response(client, query_str, url=origin.url, sort=sort)
    data_visits = response["origin"]["visits"]["nodes"]
    if sort == "DESC":
        data_visits.reverse()
    assert [
        {"visitId": x.visit} for x in storage.origin_visit_get(origin.url).results
    ] == data_visits


def test_latest_visit_type_filter(client):
    query_str = """
    query getOrigin($url: String!, $visitType: String!) {
      origin(url: $url) {
        latestVisit(visitType: $visitType) {
          visitId
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client, query_str, url=get_origins()[0].url, visitType="git"
    )
    assert data["origin"] == {"latestVisit": {"visitId": 3}}

    data, _ = utils.get_query_response(
        client, query_str, url=get_origins()[0].url, visitType="hg"
    )
    assert data["origin"] == {"latestVisit": None}


def test_latest_visit_require_snapshot_filter(client):
    query_str = """
    query getOrigin($url: String!, $requireSnapshot: Boolean!) {
      origin(url: $url) {
        latestVisit(requireSnapshot: $requireSnapshot) {
          visitId
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client,
        query_str,
        url=get_origins()[1].url,
        requireSnapshot=True,
    )
    assert data["origin"] == {"latestVisit": {"visitId": 2}}


def test_latest_visit_allowed_statuses_filter(client):
    query_str = """
    query getOrigin($url: String!, $allowedStatuses: [VisitStatusState!]!) {
      origin(url: $url) {
        latestVisit(allowedStatuses: $allowedStatuses) {
          visitId
          statuses {
            nodes {
              status
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client,
        query_str,
        url=get_origins()[1].url,
        allowedStatuses=["partial"],
    )
    assert data["origin"] == {
        "latestVisit": {"statuses": {"nodes": [{"status": "partial"}]}, "visitId": 2}
    }


@pytest.mark.parametrize("origin", get_origins())
def test_latest_snashot(client, origin):
    query_str = """
    query getOrigin($url: String!) {
      origin(url: $url) {
        latestSnapshot {
          swhid
        }
        latestVisit(requireSnapshot: true) {
          latestStatus(requireSnapshot: true) {
            snapshot {
              swhid
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client,
        query_str,
        url=origin.url,
    )
    # origin.latestSnapshot and origin.latestVisit.latestStatus.snapshot must be the same
    assert (
        data["origin"]["latestSnapshot"]
        == data["origin"]["latestVisit"]["latestStatus"]["snapshot"]
    )



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

from starlette.testclient import TestClient
import yaml

from swh.graphql import server

from . import utils
from ..data import get_snapshots


def test_valid_query(client):
    query_str = """
    query getOrigins {
      origins(first: 2) {
        nodes {
          url
        }
      }
    }
    """
    response, _ = utils.get_query_response(client, query_str)
    assert len(response["origins"]["nodes"]) == 2


def test_query_cost_simple(client):
    query_str = """
    query getOrigins {
      origins(first: 1000) {
        nodes {
          url
        }
      }
    }
    """
    errors = utils.get_error_response(client, query_str, response_code=400)
    assert (
        "The query exceeds the maximum cost of 100. Actual cost is 1000"
        in errors[0]["message"]
    )


def test_query_cost_with_no_limit(client, none_query_cost):
    # No cost validation
    query_str = """
    query getOrigins {
      origins(first: 1000) {
        nodes {
          url
        }
      }
    }
    """
    response, _ = utils.get_query_response(client, query_str)
    assert len(response["origins"]["nodes"]) == 2


def test_query_cost_origin(client):
    query_str = """
    query getOrigins {
      origins(first: 10) {
        nodes {
          url
          latestVisit {
            date
          }
          visits(first: 5) {
            nodes {
              date
              statuses {
                nodes {
                  date
                }
              }
            }
          }
          snapshots(first: 5) {
            nodes {
              swhid
            }
          }
        }
      }
    }
    """
    # Total cost here is 170
    # 10 (origin) + 10 (latestVisit) + 10*5 (visits) + 10 * 5 * 3 (status) +
    # 10 * 5*2 (snapshots) = 320
    errors = utils.get_error_response(client, query_str, response_code=400)
    assert (
        "The query exceeds the maximum cost of 100. Actual cost is 320"
        in errors[0]["message"]
    )


def test_query_cost_snapshots(client):
    query_str = """
    query getSnapshot($swhid: SWHID!) {
      snapshot(swhid: $swhid) {
        branches(first: 50) {
          nodes {
            target {
              node {
                ...on Revision {
                  swhid
                }
                ...on Directory {
                  swhid
                  entries(first: 3) {
                    nodes {
                      name {
                        text
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
    """
    # Total cost here is 157
    # 1 (snapshot) + 2 *50 (branches) + 50 * 1 (branch target)
    # + 50 * 1 (revision or Directory) +  3 * 2 = 207
    # parent multiplier is not applied when schema introspection is used
    # ie: directory entry connection cost is 3 * 2 and not 50 * 3 * 2
    errors = utils.get_error_response(
        client, query_str, swhid=str(get_snapshots()[0].swhid()), response_code=400
    )
    assert (
        "The query exceeds the maximum cost of 100. Actual cost is 207"
        in errors[0]["message"]
    )


def _test_reduced_cost_for_anonymous(client):
    # max cost is 10 for a test anonymous user
    query_str = """
    query getOrigins {
      origins(first: 11) {
        nodes {
          url
        }
      }
    }
    """
    errors = utils.get_error_response(client, query_str, response_code=400)
    assert (
        "The query exceeds the maximum cost of 10. Actual cost is 11"
        in errors[0]["message"]
    )


def test_reduced_cost_for_anonymous_user(client, anonymous_user):
    _test_reduced_cost_for_anonymous(client)


def test_reduced_cost_for_no_auth_client(mocker):
    mocker.patch.object(server, "graphql_cfg", {})
    with open(
        os.path.join(os.path.dirname(server.__file__), "config/test.yml"), "r"
    ) as config_file:
        config = yaml.load(config_file.read(), Loader=yaml.SafeLoader)
        del config["auth"]
    mocker.patch.object(server, "load_and_check_config").return_value = config
    app = server.make_app_from_configfile()
    client = TestClient(app)
    _test_reduced_cost_for_anonymous(client)



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from . import utils
from ..data import (  # get_contents,; get_directories,; get_snapshots,
    get_contents,
    get_directories,
    get_releases,
    get_revisions,
    get_snapshots,
)


def test_invalid_swhid(client):
    query_str = """
    query resolve($swhid: SWHID!) {
      resolveSWHID(swhid: $swhid) {
        __typename
      }
    }
    """
    errors = utils.get_error_response(client, query_str, swhid="swh:1:dir:invalid")
    # API will throw an error in case of an invalid SWHID
    assert len(errors) == 1
    assert "Input error: Invalid SWHID" in errors[0]["message"]


@pytest.mark.parametrize(
    "swhid",
    [
        "swh:1:rel:0949d7a8c96347dba09be8d79085b8207f345412",
        "swh:1:rev:0949d7a8c96347dba09be8d79085b8207f345412",
        "swh:1:dir:0949d7a8c96347dba09be8d79085b8207f345412",
        "swh:1:cnt:0949d7a8c96347dba09be8d79085b8207f345412",
        "swh:1:snp:0949d7a8c96347dba09be8d79085b8207f345412",
    ],
)
def test_missing_swhid(client, swhid):
    query_str = """
    query resolve($swhid: SWHID!) {
      resolveSWHID(swhid: $swhid) {
        __typename
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=swhid)
    # API will return None in case of a valid, non existing SWHID
    assert data == {"resolveSWHID": None}


@pytest.mark.parametrize("snapshot", get_snapshots())
def test_snapshot_swhid_resolve(client, snapshot):
    query_str = """
    query resolve($swhid: SWHID!) {
      resolveSWHID(swhid: $swhid) {
        __typename
        ... on Snapshot {
          swhid
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(snapshot.swhid()))
    assert data == {
        "resolveSWHID": [
            {
                "__typename": "Snapshot",
                "swhid": str(snapshot.swhid()),
            },
        ]
    }


@pytest.mark.parametrize("revision", get_revisions())
def test_revision_swhid_resolve(client, revision):
    query_str = """
    query resolve($swhid: SWHID!) {
      resolveSWHID(swhid: $swhid) {
        __typename
        ... on Revision {
          swhid
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(revision.swhid()))
    assert data == {
        "resolveSWHID": [
            {
                "__typename": "Revision",
                "swhid": str(revision.swhid()),
            }
        ]
    }


@pytest.mark.parametrize("release", get_releases())
def test_release_swhid_resolve(client, release):
    query_str = """
    query resolve($swhid: SWHID!) {
      resolveSWHID(swhid: $swhid) {
        __typename
        ... on Release {
          swhid
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(release.swhid()))
    assert data == {
        "resolveSWHID": [
            {
                "__typename": "Release",
                "swhid": str(release.swhid()),
            }
        ]
    }


@pytest.mark.parametrize("directory", get_directories())
def test_directory_swhid_resolve(client, directory):
    query_str = """
    query resolve($swhid: SWHID!) {
      resolveSWHID(swhid: $swhid) {
        __typename
        ... on Directory {
          swhid
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(directory.swhid()))
    assert data == {
        "resolveSWHID": [
            {
                "__typename": "Directory",
                "swhid": str(directory.swhid()),
            },
        ]
    }


@pytest.mark.parametrize("content", get_contents())
def test_content_swhid_resolve(client, content):
    query_str = """
    query resolve($swhid: SWHID!) {
      resolveSWHID(swhid: $swhid) {
        __typename
        ... on Content {
          swhid
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(content.swhid()))
    assert data == {
        "resolveSWHID": [
            {
                "__typename": "Content",
                "swhid": str(content.swhid()),
            },
        ]
    }



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from . import utils
from ..data import get_contents


@pytest.mark.parametrize("content", get_contents())
def test_get_content_with_hashes(client, content):
    query_str = """
    query getContentByHashes($sha1: String!, $sha256: String!,
                     $sha1_git: String!, $blake2s256: String!) {
      contentByHashes(sha1: $sha1, sha256: $sha256, sha1_git: $sha1_git,
              blake2s256: $blake2s256) {
        swhid
        id
        hashes {
          blake2s256
          sha1
          sha1_git
          sha256
        }
        length
        status
        data {
          url
        }
        mimeType {
          encoding
        }
        language {
          lang
        }
        license {
          licenses
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client,
        query_str,
        blake2s256=content.blake2s256.hex(),
        sha1=content.sha1.hex(),
        sha1_git=content.sha1_git.hex(),
        sha256=content.sha256.hex(),
    )
    archive_url = "https://archive.softwareheritage.org/api/1/"
    response = {
        "swhid": str(content.swhid()),
        "id": content.sha1_git.hex(),
        "hashes": {
            "blake2s256": content.blake2s256.hex(),
            "sha1": content.sha1.hex(),
            "sha1_git": content.sha1_git.hex(),
            "sha256": content.sha256.hex(),
        },
        "length": content.length,
        "status": content.status,
        "data": {
            "url": f"{archive_url}content/sha1:{content.sha1.hex()}/raw/",
        },
        "mimeType": None,
        "language": None,
        "license": None,
    }
    assert data["contentByHashes"] == response


@pytest.mark.parametrize("content", get_contents())
def test_get_contents_with_swhid(client, content):
    query_str = """
    query getContents($swhid: SWHID!) {
      contentsBySWHID(swhid: $swhid) {
        swhid
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(content.swhid()))
    assert data["contentsBySWHID"] == [{"swhid": str(content.swhid())}]


def test_get_contents_with_invalid_swhid(client):
    query_str = """
    query getContents($swhid: SWHID!) {
      contentsBySWHID(swhid: $swhid) {
        swhid
      }
    }
    """
    errors = utils.get_error_response(client, query_str, swhid="invalid")
    # API will throw an error in case of an invalid SWHID
    assert len(errors) == 1
    assert "Input error: Invalid SWHID" in errors[0]["message"]


def test_get_contents_with_missing_swhid(client):
    missing_sha1 = "1" * 40
    query_str = """
    query getContents($swhid: SWHID!) {
      contentsBySWHID(swhid: $swhid) {
        swhid
      }
    }
    """
    data, _ = utils.get_query_response(
        client, query_str, swhid=f"swh:1:cnt:{missing_sha1}"
    )
    assert data["contentsBySWHID"] == []


@pytest.mark.parametrize("content", get_contents())
def test_get_contents_with_hashes(client, content):
    query_str = """
    query getContents($sha1: String, $sha1_git: String, $sha256: String,
                     $blake2s256: String) {
      contentsByHashes(sha1: $sha1, sha1_git: $sha1_git, sha256: $sha256,
                      blake2s256: $blake2s256) {
        swhid
      }
    }
    """
    data, _ = utils.get_query_response(
        client,
        query_str,
        sha1=content.sha1.hex(),
        sha1_git=content.sha1_git.hex(),
        sha256=content.sha256.hex(),
        blake2s256=content.blake2s256.hex(),
    )
    assert data["contentsByHashes"] == [{"swhid": str(content.swhid())}]


@pytest.mark.parametrize("content", get_contents())
def test_get_content_with_single_hash(client, content):
    query_str = """
    query getContents($sha1: String, $sha1_git: String, $sha256: String,
                     $blake2s256: String) {
      contentsByHashes(sha1: $sha1, sha1_git: $sha1_git, sha256: $sha256,
                      blake2s256: $blake2s256) {
        swhid
      }
    }
    """
    data, _ = utils.get_query_response(
        client,
        query_str,
        sha1=content.sha1.hex(),
    )
    assert data["contentsByHashes"] == [{"swhid": str(content.swhid())}]

    data, _ = utils.get_query_response(
        client,
        query_str,
        blake2s256=content.blake2s256.hex(),
    )
    assert data["contentsByHashes"] == [{"swhid": str(content.swhid())}]


@pytest.mark.parametrize("content", get_contents())
def test_get_contents_with_one_non_matching_hash(client, content):
    query_str = """
    query getContents($sha1: String, $sha1_git: String, $sha256: String, $blake2s256: String) {
      contentsByHashes(sha1: $sha1, sha1_git: $sha1_git, sha256: $sha256,
                      blake2s256: $blake2s256) {
        swhid
      }
    }
    """
    data, _ = utils.get_query_response(
        client,
        query_str,
        obj_type="contentsByHashes",
        sha1=content.sha1.hex(),
        sha1_git="a" * 20,  # hash is valid, but not matching the object
    )
    assert data["contentsByHashes"] == []


def test_get_content_with_invalid_hashes(client):
    content = get_contents()[0]
    query_str = """
    query getContents($sha1: String, $sha1_git: String, $sha256: String,
                      $blake2s256: String) {
      contentsByHashes(sha1: $sha1, sha1_git: $sha1_git, sha256: $sha256,
                       blake2s256: $blake2s256) {
        swhid
      }
    }
    """
    errors = utils.get_error_response(
        client,
        query_str,
        sha1="invalid",  # Only one hash is invalid
        sha1_git=content.sha1_git.hex(),
        sha256=content.sha256.hex(),
    )
    # API will throw an error in case of an invalid content hash
    assert len(errors) == 1
    assert "Input error: Invalid content hash" in errors[0]["message"]


def test_get_content_with_no_hashes(client):
    query_str = """
    query getContents($sha1: String, $sha1_git: String, $sha256: String, $blake2s256: String) {
      contentsByHashes(sha1: $sha1, sha1_git: $sha1_git, sha256: $sha256,
                      blake2s256: $blake2s256) {
        swhid
      }
    }
    """
    errors = utils.get_error_response(
        client,
        query_str,
    )
    assert len(errors) == 1
    assert (
        "Input error: At least one of the four hashes must be provided"
        in errors[0]["message"]
    )


def test_get_content_as_target(client):
    # SWHID of a test dir with a file entry
    directory_swhid = "swh:1:dir:87b339104f7dc2a8163dec988445e3987995545f"
    query_str = """
    query getDirectory($swhid: SWHID!) {
      directory(swhid: $swhid) {
        swhid
        entries(first: 2) {
          nodes {
            target {
              type
              node {
                ...on Content {
                  swhid
                  length
                }
              }
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=directory_swhid)
    content_obj = data["directory"]["entries"]["nodes"][1]["target"]["node"]
    assert content_obj == {
        "length": 4,
        "swhid": "swh:1:cnt:86bc6b377e9d25f9d26777a4a28d08e63e7c5779",
    }



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import List

from swh.graphql.utils import utils
from swh.model.model import DirectoryEntry, ObjectType, Sha1Git

from .base_connection import BaseConnection, ConnectionData
from .base_node import BaseNode


class BaseDirectoryEntryNode(BaseNode):
    def target_hash(self) -> Sha1Git:
        assert self._node is not None
        return self._node.target

    def target_type(self) -> ObjectType:
        mapping = {
            "file": ObjectType.CONTENT,
            "dir": ObjectType.DIRECTORY,
            "rev": ObjectType.REVISION,
        }
        assert self._node is not None
        return mapping[self._node.type]


class DirectoryEntryNode(BaseDirectoryEntryNode):
    """
    Node resolver for a directory entry requested with a
    directory SWHID and a relative path
    """

    def _get_node_data(self):
        # STORAGE-TODO, archive is returning a dict
        # return DirectoryEntry object instead
        return self.archive.get_directory_entry_by_path(
            directory_id=self.kwargs.get("directorySWHID").object_id,
            path=self.kwargs.get("path"),
        )


class DirectoryEntryConnection(BaseConnection):
    """
    Connection resolver for entries in a directory
    """

    from .directory import BaseDirectoryNode

    obj: BaseDirectoryNode

    _node_class = BaseDirectoryEntryNode

    def _get_connection_data(self) -> ConnectionData:
        # FIXME, using dummy(local) pagination, move pagination to backend
        # STORAGE-TODO
        entries: List[DirectoryEntry] = self.archive.get_directory_entries(
            self.obj.swhid.object_id
        ).results
        name_include = self.kwargs.get("nameInclude")
        if name_include is not None:
            # STORAGE-TODO, move this filter to swh-storage
            entries = [
                x
                for x in entries
                if name_include.casefold() in x.name.decode().casefold()
            ]
        return utils.get_local_paginated_data(
            entries, self._get_first_arg(), self._get_after_arg()
        )
